{***********************************************************************
Unit ColorWheel.PAS v1.0 Build 0199
    (c) 1999 by Andreas Moser, AMoser@t-online.de

    Delphi version : Delphi 3

    ColorWheel.pas is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    which are not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

********************************************************************************}

unit ColorWheel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, gfx_basedef;

type
  TColorWheel = class(TPanel)
  private
    fColImage:TImage;
    FFillColor:TColor;
    dx,dy,sdX,sdY,sDegr,lCX,lCY,iX,iY,XPos,YPos:Integer;
    fActRed,fActGreen,fActBlue:Byte;
    FOnChange:TnotifyEvent;
    procedure DrawRectangle;
    procedure DrawCircle;
    procedure ImageMouseUp(Sender: TObject; Button: TMouseButton;
    Shift: TShiftState; X, Y: Integer);
    procedure ImageMouseMove(Sender: TObject;
    Shift: TShiftState; X, Y: Integer);
    Procedure SetColor(Col:Tcolor);
    Function GetColor:TColor;
  protected
    { Protected-Deklarationen }
  public
    { Public-Deklarationen }
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
  published
    property FillColor:Tcolor read GetColor write SetColor;
    property ActRed:Byte read fActred;
    property ActGreen:Byte read fActGreen;
    property ActBlue:Byte read fActBlue;
    property OnChange:TNotifyEvent read fOnChange write fOnChange;
    { Published-Deklarationen }
  end;

procedure Register;

implementation

constructor TColorWheel.Create;
var i,s,v,x,y,x2,y2:Integer;
    RGBTrip:TRGBTriple;
    HSVTrip:THSVTriple;
begin
  inherited Create(AOwner);
  fOnChange:=nil;
  DoubleBuffered:=true;
  Width:=204;
  Height:=204;
  FColImage:=TImage.Create(self);
  FColImage.parent:=self;
  FColImage.Width:=Width-4;
  FColImage.Height:=Height-4;
  FColImage.Left:=2;
  FColImage.Top:=2;
  FFillColor:=ClBtnFace;
  FColImage.Picture.Bitmap.PixelFormat:=pf24Bit;
  FColImage.Picture.Bitmap.Width:=Width;
  FColImage.picture.Bitmap.Height:=Height;
  FColImage.Picture.Bitmap.Canvas.Brush.Color:=FFillColor;
  FColImage.Picture.Bitmap.Canvas.FillRect(Rect(0,0,Width,Height));
  FColImage.OnMouseUp:=ImageMouseUp;
  FColImage.OnMouseMove:=ImageMouseMove;

  dx:=FColImage.Width div 2;
  dy:=FColImage.Height div 2;
  FColImage.Picture.Bitmap.Canvas.Pen.Color:=clBlack;
  FColImage.Picture.Bitmap.Canvas.Pen.Width :=3;
  sDx:=FColImage.Left+FcolImage.Width div 2- 7 div 2;
  sDy:=FColImage.Top+13-7 div 2;
  sDegr:=0;
  DrawCircle;
  DrawRectangle;
  XPos:=lCX+1;
  YPos:=lCY+1;
end;

destructor TColorWheel.Destroy;
begin
  inherited Destroy;
end;

procedure TColorWheel.DrawCircle;
var i,s,v,x,y,x2,y2,pw:Integer;
    RGBTrip:TRGBTriple;
    HSVTrip:THSVTriple;
    Row:pRGBArray;
    RGBT:TRGBTriple;
begin
FColImage.Picture.Bitmap.Canvas.MoveTo(Round(dx+(dx-25)*Sin(0*PI/180)),Round(dy+(dy-25)*Cos(0*PI/180)));
HSVTrip.hsvSaturation:=255;
HSVTrip.hsvValue:=255;
for i:=1 to 360 do
    begin
      HSVTrip.hsvHue:=i;
      HSVtoRGB(HSVTrip,RGBTrip);
      FColImage.Picture.Bitmap.Canvas.Pen.Color:=RgbToColor(RGBTrip);
      x:=Round((dx-25)*Sin((i)*PI/180));
      y:=Round((dy-25)*Cos((i)*PI/180));
      x2:=Round((dx-5)*Sin((i)*PI/180));
      y2:=Round((dy-5)*Cos((i)*PI/180));
      FColImage.Picture.Bitmap.Canvas.MoveTo(dx+x,dy+y);
      FColImage.Picture.Bitmap.Canvas.LineTo(dx+x2,dy+y2);
    end;
Row:=fColImage.Picture.Bitmap.Scanline[Sdy];
RGBT:=Row[Sdx];
RGBT.rgbtBlue:=255-RGBT.rgbtBlue;
RGBT.rgbtGreen:=255-RGBT.rgbtGreen;
RGBT.rgbtRed:=255-RGBT.rgbtRed;
FColImage.Picture.Bitmap.Canvas.Pen.Color:=RGBToColor(RGBT);
pw:=FColImage.Picture.Bitmap.Canvas.Pen.Width;
FColImage.Picture.Bitmap.Canvas.Pen.Width:=1;
FColImage.Picture.Bitmap.Canvas.Brush.Style:=bsclear;
FColImage.Picture.Bitmap.Canvas.Ellipse(sdx-3,sdy-3,sdx+3,sdy+3);
FColImage.Picture.Bitmap.Canvas.Pen.Width:=pw;
FColImage.Picture.Bitmap.Canvas.Brush.Style:=bsSolid;
end;

Procedure TColorWheel.SetColor(Col:Tcolor);
begin
FFillColor:=col;
FColImage.Picture.Bitmap.Canvas.Brush.Color:=Col;
FColImage.Picture.Bitmap.Canvas.FillRect(Rect(0,0,Width,Height));
DrawCircle;
DrawRectangle;
invalidate;
end;

Function TColorWheel.GetColor:TColor;
begin
result:=FFillColor;
end;

procedure TColorWheel.DrawRectangle;
var i,Col,pw:Integer;
    x3,x4,y1,y4:Integer;
    RGBTrip:TRGBTriple;
    Row:pRGBArray;
    HSV:THSVTriple;
    RGBT:TRGBTriple;
begin
//  dx:=FColImage.Picture.Bitmap.Width div 2;
//  dy:=FColImage.Picture.Bitmap.Height div 2;
  lCX:=Round(dx+(dx-30)*Sin((315)*PI/180));
  y1:=Round(dy+(dy-30)*Cos((315)*PI/180));

//  x2:=Round(dx+(dx-30)*Sin((45)*PI/180));
//  y2:=Round(dy+(dy-30)*Cos((45)*PI/180));

  x3:=Round(dx+(dx-30)*Sin((135)*PI/180));
//  y3:=Round(dy+(dy-30)*Cos((135)*PI/180));

  x4:=Round(dx+(dx-30)*Sin((225)*PI/180));
  y4:=Round(dy+(dy-30)*Cos((225)*PI/180));
  lCY:=y4;

  iX:=(x3-x4);
  iY:=(y1-y4);
for i:=0 to iX {255} do
 begin
  Row:=FColImage.Picture.Bitmap.ScanLine[i+x4];
  for Col:=0 to iY{255} do
      begin
      RgbTrip.rgbtRed:=TrimReal(0,255,(255-(i*(255/iX))+((i*Col*(255/iX))/ iX)));
      RgbTrip.rgbtBlue:=TrimReal(0,255,(255/iX)*Col);
      RgbTrip.rgbtGreen:=TrimReal(0,255,(255/iX)*Col);
      RGBtoHSV(RGBTrip,hsv);
      HSV.hsvHue:=sDegr;
      HSV.hsvHue:=LoopInt(0,359,HSV.hsvHue);
      HSVtoRGB(HSV,RGBTrip);
      Row[Col+y4].rgbtRed :=RGBTrip.rgbtRed;
      Row[Col+y4].rgbtBlue :=RGBTrip.rgbtBlue;
      Row[Col+y4].rgbtGreen :=RGBTrip.rgbtGreen;
      end;
 end;
FColImage.Repaint;
Row:=fColImage.Picture.Bitmap.Scanline[YPos];
fActRed:=Row[XPos].rgbtRed;
fActGreen:=Row[XPos].rgbtGreen;
fActBlue:=Row[XPos].rgbtBlue;

RGBT.rgbtBlue:=255-fActBlue;
RGBT.rgbtGreen:=255-fActGreen;
RGBT.rgbtRed:=255-fActRed;

FColImage.Picture.Bitmap.Canvas.Pen.Color:=RGBToColor(RGBT);
pw:=FColImage.Picture.Bitmap.Canvas.Pen.Width;
FColImage.Picture.Bitmap.Canvas.Pen.Width:=1;
FColImage.Picture.Bitmap.Canvas.Brush.Style:=bsclear;
FColImage.Picture.Bitmap.Canvas.Ellipse(xpos-3,ypos-3,xpos+3,ypos+3);
FColImage.Picture.Bitmap.Canvas.Pen.Width:=pw;
FColImage.Picture.Bitmap.Canvas.Brush.Style:=bsSolid;
end;


procedure TColorWheel.ImageMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var dx,dy,r,a,b,pw:Integer;
    sinalpha,cosalpha:double;
    Row:pRGBArray;
    RGBT:TRGBTriple;
begin
   dx:=FcolImage.Width div 2;
   dy:=FcolImage.Height div 2;
   a:=Abs(X-dx);
   b:=Abs(y-dy);
 {  d:=MaxInt2(a,b);
   b:=MinInt2(a,b);
   a:=d;
  } r:=Round(sqrt((a*a)+(b*b)));
   sinalpha:=a/r;
   cosAlpha:=b/r;
   if (r>dx-25) and (r<dx-5) then  begin
      if (x>dx) and (y<dy) then
      begin
        Sdx:=Round(dx+(dx-15)*(SinAlpha))+ 7 div 2;
        Sdy:=Round(dy-(dy-15)*(CosAlpha))+7 div 2;
        sDegr:=180-Round(xArcSin(SinAlpha)/pi*180);
      end else
      if (x>dx) and (y>dy) then
      begin
        Sdx:=Round(dx+(dx-15)*(SinAlpha))+7 div 2;
        Sdy:=Round(dy+(dy-15)*(CosAlpha))+7 div 2;
        sDegr:=Round(xArcSin(SinAlpha)/pi*180);
      end else
      if (x<dx) and (y>dy) then
      begin
        Sdx:=Round(dx-(dx-15)*(SinAlpha))+7 div 2;
        Sdy:=Round(dy+(dy-15)*(CosAlpha))+7 div 2;
        sDegr:=360-Round(xArcSin(SinAlpha)/pi*180);
      end else
      if (x<dx) and (y<dy) then
      begin
        Sdx:=Round(dx-(dx-15)*(SinAlpha))+7 div 2;
        Sdy:=Round(dy-(dy-15)*(CosAlpha))+7 div 2;
        sDegr:=180+Round(xArcSin(SinAlpha)/pi*180);
      end;
      DrawCircle;
      DrawRectangle;
      if Assigned(fOnChange) then OnChange(self);
   end;
end;

procedure TColorWheel.ImageMouseMove(Sender: TObject;
    Shift: TShiftState; X, Y: Integer);
var i,Col:Integer;
    x3,x4,y1,y4,pw:Integer;
    RGBTrip:TRGBTriple;
    Row:pRGBArray;
    HSV:THSVTriple;
    RGBT:TRGBTriple;
begin
if not (ssLeft in Shift) then exit;
ImageMouseUp(Sender,mbleft,Shift, X, Y);
//  dx:=FColImage.Picture.Bitmap.Width div 2;
//  dy:=FColImage.Picture.Bitmap.Height div 2;
  lCX:=Round(dx+(dx-30)*Sin((315)*PI/180));
  y1:=Round(dy+(dy-30)*Cos((315)*PI/180));

//  x2:=Round(dx+(dx-30)*Sin((45)*PI/180));
//  y2:=Round(dy+(dy-30)*Cos((45)*PI/180));

  x3:=Round(dx+(dx-30)*Sin((135)*PI/180));
//  y3:=Round(dy+(dy-30)*Cos((135)*PI/180));

  x4:=Round(dx+(dx-30)*Sin((225)*PI/180));
  y4:=Round(dy+(dy-30)*Cos((225)*PI/180));
  lCY:=y4;

  iX:=(x3-x4);
  iY:=(y1-y4);
if (X>=y4) and (Y>=x4) and (X<=y4+iY) and (Y<=x4+iX) then
 begin
 FColImage.Picture.Bitmap.Canvas.Brush.Color:=FFillColor;
 FColImage.Picture.Bitmap.Canvas.FillRect(Rect(y4-3,x4-3,y4+iy+3,x4+ix+3));
 for i:=0 to iX  do
  begin
  Row:=FColImage.Picture.Bitmap.ScanLine[i+x4];
  for Col:=0 to iY do
   begin
   RgbTrip.rgbtRed:=TrimReal(0,255,(255-(i*(255/iX))+((i*Col*(255/iX))/ iX)));
   RgbTrip.rgbtBlue:=TrimReal(0,255,(255/iX)*Col);
   RgbTrip.rgbtGreen:=TrimReal(0,255,(255/iX)*Col);
   RGBtoHSV(RGBTrip,hsv);
   HSV.hsvHue:=sDegr;
   HSV.hsvHue:=LoopInt(0,359,HSV.hsvHue);
   HSVtoRGB(HSV,RGBTrip);
   Row[Col+y4].rgbtRed :=RGBTrip.rgbtRed;
   Row[Col+y4].rgbtBlue :=RGBTrip.rgbtBlue;
   Row[Col+y4].rgbtGreen :=RGBTrip.rgbtGreen;
   end;
  end;
 FColImage.Repaint;
 Row:=fColImage.Picture.Bitmap.Scanline[y];
 RGBT:=Row[x];
 XPos:=x;
 Ypos:=y;
 fActRed:=RGBT.rgbtRed;
 fActGreen:=RGBT.rgbtGreen;
 fActBlue:=RGBT.rgbtBlue;
 RGBT.rgbtBlue:=255-RGBT.rgbtBlue;
 RGBT.rgbtGreen:=255-RGBT.rgbtGreen;
 RGBT.rgbtRed:=255-RGBT.rgbtRed;

 FColImage.Picture.Bitmap.Canvas.Pen.Color:=RGBToColor(RGBT);
 pw:=FColImage.Picture.Bitmap.Canvas.Pen.Width;
 FColImage.Picture.Bitmap.Canvas.Pen.Width:=1;
 FColImage.Picture.Bitmap.Canvas.Brush.Style:=bsclear;
 FColImage.Picture.Bitmap.Canvas.Ellipse(x-3,y-3,x+3,y+3);
 FColImage.Picture.Bitmap.Canvas.Pen.Width:=pw;
 FColImage.Picture.Bitmap.Canvas.Brush.Style:=bsSolid;
 if Assigned(fOnChange) then OnChange(self);
 end;
end;



procedure Register;
begin
  RegisterComponents('AM FX Pro', [TColorWheel]);
end;

end.
