unit gfx_colorpalette;
{***********************************************************************
Unit gfx_colorpalette.PAS v1.5 0801
    (c) by Andreas Moser, amoser@amoser.de

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_colorpalette is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    This unit of gfx_lib is, in part, based on the work of Anders Melander, anders@melander.dk
    (especially colorlookup and some palette functions)
    Some algorythms are based on or taken from the sources written by Earl F. Glynn, efg's computer lab

********************************************************************************}

interface
uses gfx_basedef,
     gfx_syswrapper,
     gfx_colorquantizer,
     gfx_dib;

 type
  TInverseLookup = array[0..1 SHL 15-1] of SmallInt;
  PInverseLookup = ^TInverseLookup;

  PalArray =  array[byte] of TxPaletteEntry;
  PPalArray = ^PalArray;

type
{  TColorLookup = class(TObject)
  private
    FColors		: integer;
  public
    constructor Create(Palette: hxPalette); virtual;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; virtual; abstract;
    property Colors: integer read FColors;
  end;


  TFastColorLookup = class(TColorLookup)
  private
    FPaletteEntries	: PPalArray;
    FInverseLookup	: PInverseLookup;
  public
    constructor Create(Palette: hxPalette); override;
    destructor Destroy; override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; override;
  end;

  // TSlowColorLookup implements a precise but very slow generic color mapper.
  // It uses the GetNearestPaletteIndex GDI function.
  // Note: Tests has shown TFastColorLookup to be more precise than
  // TSlowColorLookup in many cases. I can't explain why...
  TSlowColorLookup = class(TColorLookup)
  private
    FPaletteEntries	: PPalArray;
    FPalette		: hxPalette;
  public
    constructor Create(Palette: hxPalette); override;
    destructor Destroy; override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; override;
  end;

  // TNetscapeColorLookup maps colors to the netscape 6*6*6 color cube.
  TNetscapeColorLookup = class(TColorLookup)
  public
    constructor Create(Palette: hxPalette); override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; override;
  end;

  // TGrayWindowsLookup maps colors to 4 shade palette.
  TGrayWindowsLookup = class(TSlowColorLookup)
  public
    constructor Create(Palette: hxPalette); override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; override;
  end;

  // TGrayScaleLookup maps colors to a uniform 256 shade palette.
  TGrayScaleLookup = class(TColorLookup)
  public
    constructor Create(Palette: hxPalette); override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; override;
  end;

  // TMonochromeLookup maps colors to a black/white palette.
  TMonochromeLookup = class(TColorLookup)
  public
    constructor Create(Palette: hxPalette); override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char; override;
  end;
 }
  TColorLookup = class(TObject)
  private
   FPaletteType:TColorReduction;
   FColors		: integer;
   FPaletteEntries	: pPaletteEntries;
   FInverseLookup	: PInverseLookup;
  public
    constructor Create(Palette: HxPalette;PaletteType:TColorReduction);virtual;
    destructor Destroy;override;
    function Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
    property Colors: integer read FColors;
  end;

  type
  TGetBitmapCallback = procedure (const Counter:  INTEGER;
                                  var OK:  BOOLEAN; var Bitmap:  TXBitmap);

  function  MakePalette(PaletteType:TColorReduction):HxPalette;
  function  CreateOptimizedPaletteforSingleBitmap(const Bitmap:  TxBitmap;
                                                  const ColorBits:  INTEGER): HxPalette;
  function  CreateOptimizedPaletteforManyBitmaps(const CallBack:  TGetBitmapCallback;
                                                 const ColorBits:  INTEGER): HxPalette;
  function doCreateOptimizedPaletteFromSingleBitmap(const DIB: TDIBReader;Colors, ColorBits: integer; Windows: boolean): HxPalette;
  function CreateOptimizedPaletteFromSingleBitmap(const Bitmap: TXBitmap;Colors, ColorBits: integer; Windows: boolean): HxPalette;

  function  ReadPaletteFromFile(FileName:String):HxPalette;
  procedure ApplyPaletteToGraphic(Palette:HxPalette;Graphic:TXGraphic;ChangePixelformat:Boolean;NewPixelformat:TXPixelformat);
  procedure ApplyPaletteToDC(DC:HxDC;Palette:HxPalette);


  procedure LogicalPaletteToRGBQuad(const Start         :  WORD;
                                    const Count         :  WORD;
                                    var   LogicalPalette:  TPaletteEntries;
                                    var   RGBQuad       :  TRGBQuadarray);


  procedure WMQueryNewPalette(formcanvashandle:TXHandle; PaletteHandle:  HxPalette;
                              var Msg:  TXWMQueryNewPalette);

  procedure WMPaletteChanged(formcanvashandle:TXHandle;formhandle:TXHandle; PaletteHandle:  HxPalette;
                            var Msg:  TXWMPaletteChanged);


implementation
{
constructor TColorLookup.Create(Palette: hxPalette);
begin
  inherited Create;
end;

constructor TFastColorLookup.Create(Palette: hxPalette);
var
  i			: integer;
  InverseIndex		: integer;
begin
  inherited Create(Palette);

  GetMem(FPaletteEntries, sizeof(TxPaletteEntry) * 256);
  FColors := xGetPaletteEntries(Palette, 0, 256, FPaletteEntries^);

  New(FInverseLookup);
  for i := low(TInverseLookup) to high(TInverseLookup) do
    FInverseLookup^[i] := -1;

  // Premap palette colors
  if (FColors > 0) then
    for i := 0 to FColors-1 do
      with FPaletteEntries^[i] do
      begin
        InverseIndex := (peRed SHR 3) OR ((peGreen AND $F8) SHL 2) OR ((peBlue AND $F8) SHL 7);
        if (FInverseLookup^[InverseIndex] = -1) then
          FInverseLookup^[InverseIndex] := i;
      end;
end;

destructor TFastColorLookup.Destroy;
begin
  if (FPaletteEntries <> nil) then
    FreeMem(FPaletteEntries);
  if (FInverseLookup <> nil) then
    Dispose(FInverseLookup);

  inherited Destroy;
end;

// Map color to arbitrary palette
function TFastColorLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
var
  i			: integer;
  InverseIndex		: integer;
  Delta			,
  MinDelta		,
  MinColor		: integer;
begin
  // Reduce color space with 3 bits in each dimension
  InverseIndex := (Red SHR 3) OR ((Green AND $F8) SHL 2) OR ((Blue AND $F8) SHL 7);

  if (FInverseLookup^[InverseIndex] <> -1) then
    Result := char(FInverseLookup^[InverseIndex])
  else
  begin
    // Sequential scan for nearest color to minimize euclidian distance
    MinDelta := 3 * (256 * 256);
    MinColor := 0;
    for i := 0 to FColors-1 do
      with FPaletteEntries[i] do
      begin
        Delta := ABS(peRed - Red) + ABS(peGreen - Green) + ABS(peBlue - Blue);
        if (Delta < MinDelta) then
        begin
          MinDelta := Delta;
          MinColor := i;
        end;
      end;
    Result := char(MinColor);
    FInverseLookup^[InverseIndex] := MinColor;
  end;

  with FPaletteEntries^[ord(Result)] do
  begin
    R := peRed;
    G := peGreen;
    B := peBlue;
  end;
end;

constructor TSlowColorLookup.Create(Palette: hxPalette);
begin
  inherited Create(Palette);
  FPalette := Palette;
  FColors := xGetPaletteEntries(Palette, 0, 256, nil^);
  if (FColors > 0) then
  begin
    GetMem(FPaletteEntries, sizeof(TxPaletteEntry) * FColors);
    FColors := xGetPaletteEntries(Palette, 0, 256, FPaletteEntries^);
  end;
end;

destructor TSlowColorLookup.Destroy;
begin
  if (FPaletteEntries <> nil) then
    FreeMem(FPaletteEntries);

  inherited Destroy;
end;

// Map color to arbitrary palette
function TSlowColorLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
begin
  Result := char(GetNearestPaletteIndex(FPalette, Red OR (Green SHL 8) OR (Blue SHL 16)));
  if (FPaletteEntries <> nil) then
    with FPaletteEntries^[ord(Result)] do
    begin
      R := peRed;
      G := peGreen;
      B := peBlue;
    end;
end;

constructor TNetscapeColorLookup.Create(Palette: hxPalette);
begin
  inherited Create(Palette);
  FColors := 6*6*6; // This better be true or something is wrong
end;

// Map color to netscape 6*6*6 color cube
function TNetscapeColorLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
begin
  R := (Red+3) DIV 51;
  G := (Green+3) DIV 51;
  B := (Blue+3) DIV 51;
  Result := char(B + 6*G + 36*R);
  R := R * 51;
  G := G * 51;
  B := B * 51;
end;

constructor TGrayWindowsLookup.Create(Palette: hxPalette);
begin
  inherited Create(Palette);
  FColors := 4;
end;

// Convert color to windows grays
function TGrayWindowsLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
begin
  Result := inherited Lookup(xMulDiv(Red, 77, 256),
    xMulDiv(Green, 150, 256), xMulDiv(Blue, 29, 256), R, G, B);
end;

constructor TGrayScaleLookup.Create(Palette: hxPalette);
begin
  inherited Create(Palette);
  FColors := 256;
end;

// Convert color to grayscale
function TGrayScaleLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
begin
  Result := char((Blue*29 + Green*150 + Red*77) DIV 256);
  R := ord(Result);
  G := ord(Result);
  B := ord(Result);
end;

constructor TMonochromeLookup.Create(Palette: hxPalette);
begin
  inherited Create(Palette);
  FColors := 2;
end;

// Convert color to black/white
function TMonochromeLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
begin
  if ((Blue*29 + Green*150 + Red*77) > 32512) then
  begin
    Result := #1;
    R := 255;
    G := 255;
    B := 255;
  end else
  begin
    Result := #0;
    R := 0;
    G := 0;
    B := 0;
  end;
end;
}
// -----------------------------------------------------------------------------
//
//			TColorLookup
//
// -----------------------------------------------------------------------------

constructor TColorLookup.Create(Palette: HxPalette;PaletteType:TColorReduction);
var i:Integer;
    InverseIndex		: integer;
begin
  inherited Create;
  FPaletteType:=PaletteType;
  case FPaletteType of
   rmGrayScale:  FColors := 256;
   rmMonochrome: FColors := 2;
   rmNetscape:   FColors := 6*6*6; // This better be true or something is wrong
   else
            begin

               GetMem(FPaletteEntries, sizeof(TXPaletteEntry) * 256);
               FColors := xGetPaletteEntries(Palette, 0, 256, FPaletteEntries^);

               New(FInverseLookup);
               for i := low(TInverseLookup) to high(TInverseLookup) do FInverseLookup^[i] := -1;

               // Premap palette colors
               if (FColors > 0) then
               for i := 0 to FColors-1 do
               with FPaletteEntries^[i] do
               begin
                 InverseIndex := (peRed SHR 3) OR ((peGreen AND $F8) SHL 2) OR ((peBlue AND $F8) SHL 7);
                 if (FInverseLookup^[InverseIndex] = -1) then FInverseLookup^[InverseIndex] := i;
               end;
            end;
   end;
end;

destructor TColorLookup.Destroy;
begin
  if (FPaletteEntries <> nil) then FreeMem(FPaletteEntries);
  if (FInverseLookup <> nil) then Dispose(FInverseLookup);
  inherited Destroy;
end;

function TColorLookup.Lookup(Red, Green, Blue: BYTE; var R, G, B: BYTE): char;
var
  i			: integer;
  InverseIndex		: integer;
  Delta			,
  MinDelta		,
  MinColor		: integer;
begin
  case FPaletteType of
    rmNetscape:begin
               R := (Red+3) DIV 51;
               G := (Green+3) DIV 51;
               B := (Blue+3) DIV 51;
               Result := char(B + 6*G + 36*R);
               R := R * 51;
               G := G * 51;
               B := B * 51;
               end;
    rmMonochrome:
               begin
               if ((Blue*29 + Green*150 + Red*77) > 32512) then
               begin
                 Result := #1;
                 R := 255;
                 G := 255;
                 B := 255;
               end else
               begin
                 Result := #0;
                 R := 0;
                 G := 0;
                 B := 0;
               end;
               end;
    rmGrayScale:begin
                FColors := 256;
                Result := char((Blue*29 + Green*150 + Red*77) DIV 256);
                R := ord(Result);
                G := ord(Result);
                B := ord(Result);
                end;
    else
               begin
                 InverseIndex := (Red SHR 3) OR ((Green AND $F8) SHL 2) OR ((Blue AND $F8) SHL 7);
                 if (FInverseLookup^[InverseIndex] <> -1) then
                 Result := char(FInverseLookup^[InverseIndex])
                 else
                 begin
                   // Sequential scan for nearest color to minimize euclidian distance
                   MinDelta := 3 * (256 * 256);
                   MinColor := 0;
                   for i := 0 to FColors-1 do
                   with FPaletteEntries[i] do
                   begin
                     Delta := ABS(peRed - Red) + ABS(peGreen - Green) + ABS(peBlue - Blue);
                     if (Delta < MinDelta) then
                     begin
                       MinDelta := Delta;
                       MinColor := i;
                     end;
                   end;
                   Result := char(MinColor);
                   FInverseLookup^[InverseIndex] := MinColor;
                 end;
                 with FPaletteEntries^[ord(Result)] do
                 begin
                   R := peRed;
                   G := peGreen;
                   B := peBlue;
                 end;
               end;
    end;
end;


// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------
//
//			Palette generators
//
// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------


// -----------------------------------------------------------------------------
//
//			MakePalette
//
// -----------------------------------------------------------------------------
function MakePalette(PaletteType:TColorReduction):HxPalette;
var
  MaxLogPalette:TxMaxLogPalette;
  i                     :Integer;
  r,g,b                 :Word;
  DC			:HxDC;
const
  Values2		:array[0..1] of byte = (0, 255);
  Values4               :array[0..3] of byte = (0, 128, 192, 255);

begin
  Result:=0;
  MaxLogPalette.palVersion:=PaletteVersion;
  case PaletteType of
  rmGrayScale:begin
              MaxLogPalette.palNumEntries:=256;
              for i:=0 To 255 Do
              begin
                MaxLogPalette.palPalEntry[i].peRed    :=i;//*63 Div 255;
                MaxLogPalette.palPalEntry[i].peGreen  :=i;//*63 Div 255;
                MaxLogPalette.palPalEntry[i].peBlue   :=i;//*63 Div 255;
                MaxLogPalette.palPalEntry[i].peFlags  :=xPC_NOCOLLAPSE;
              end;
              Result := xCreatePalette(pXLogPalette(@MaxLogPalette)^);
              end;
  rmStandard: begin
              MaxLogPalette.palNumEntries:=256;
              i:=0;
              for R:=0 to 7 do
              begin
                for G:=0 to 7 do
                begin
                  for B:=0 to 3 do
                  begin
                     MaxLogPalette.palPalEntry[i].peRed    :=(R+1)*8-1;
                     MaxLogPalette.palPalEntry[i].peGreen  :=(G+1)*8-1;
                     MaxLogPalette.palPalEntry[i].peBlue   :=(B+1)*16-1;
                     MaxLogPalette.palPalEntry[i].peFlags  :=xPC_NOCOLLAPSE;
                     Inc(i);
                  end;
                end;
              end;
              Result := xCreatePalette(pxLogPalette(@MaxLogPalette)^);
              end;
    rmWindows256:
              begin
              DC := xGetDC(0);
              try
                Result := xCreateHalfTonePalette(DC);
              finally
                xReleaseDC(0, DC);
              end;
              end;
    rmMonochrome:
              begin
              MaxLogPalette.palNumEntries := 2;
              for i := 0 to 1 do
              begin
                with (MaxLogPalette.palPalEntry[i]) do
                begin
                  peRed := Values2[i];
                  peGreen := Values2[i];
                  peBlue  := Values2[i];
                  peFlags := xPC_NOCOLLAPSE;
                end;
              end;
              Result := xCreatePalette(pxLogPalette(@MaxLogPalette)^);
              end;
    rmWindowsGray:
              begin
              MaxLogPalette.palNumEntries := 4;
              for i := 0 to 3 do
              begin
                with (MaxLogPalette.palPalEntry[i]) do
                begin
                  peRed := Values4[i];
                  peGreen := Values4[i];
                  peBlue  := Values4[i];
                  peFlags := xPC_NOCOLLAPSE;
                end;
              end;
              Result := xCreatePalette(pxLogPalette(@MaxLogPalette)^);
              end;
    rmDesktop:begin
              MaxLogPalette.palNumEntries := 256;
              DC := xGetDC(0);
              TRY
                i :=  xGetSystemPaletteEntries(DC,0, 255, MaxLogPalette.palPalEntry)
              FINALLY
                xReleaseDC(0, DC)
              end;
              if i <>0 then Result := xCreatePalette(pxLogPalette(@MaxLogPalette)^)
              end;
    rmWindows20:begin
                result:=xGetStockObject(xDEFAULT_PALETTE);
               end;
    end;


end;

// -----------------------------------------------------------------------------
//
//			CreateOptimizedPaletteForSingleBitmap
//
// -----------------------------------------------------------------------------
  // Adapted from MSJ "Wicked Code" article, Oct 97, pp. 79-84
  // Bitmap must be a DIB.
function  CreateOptimizedPaletteforSingleBitmap(const Bitmap:  TXBitmap; const ColorBits:  INTEGER): hxPalette;
var
      ColorQuantizer     :  TColorQuantizer;
      ScreendeviceContext:  hxDC;
      i                  :  INTEGER;
      LogicalPalette     :  TxMaxLogPalette;
      RGBQuadarray       :  TRGBQuadarray;
begin
    LogicalPalette.palVersion    := PaletteVersion;
    LogicalPalette.palNumEntries := 256;

    ScreendeviceContext := xGetDC(0);
    TRY
      xGetSystemPaletteEntries(ScreendeviceContext,
                                0,256, LogicalPalette.palPalEntry[0]);   // Need first 10 and last 10
    FINALLY
      xReleaseDC(0, ScreendeviceContext)
    end;

    // Normally for 24-bit images, use ColorBits of 5 or 6.  for 8-bit images
    // use ColorBits = 8.
    ColorQuantizer := TColorQuantizer.Create(236, ColorBits);
    TRY
      ColorQuantizer.ProcessImage(Bitmap.Handle);
      ColorQuantizer.GetColorTable(RGBQuadarray);

      // Skip over first 10 and last 10 "fixed" entries
      for i := 0 TO 255-20 DO
      begin
        // Skip over first 10 "fixed" entries in system palette
        LogicalPalette.palPalEntry[10+i].peRed   := RGBQuadarray[i].rgbRed;
        LogicalPalette.palPalEntry[10+i].peGreen := RGBQuadarray[i].rgbGreen;
        LogicalPalette.palPalEntry[10+i].peBlue  := RGBQuadarray[i].rgbBlue;
        LogicalPalette.palPalEntry[10+i].peFlags := RGBQuadarray[i].rgbReserved
      end;
      RESULT := xCreatePalette(pxLogPalette(@LogicalPalette)^)
    FINALLY
      ColorQuantizer.Free
    end
end;


// -----------------------------------------------------------------------------
//
//			CreateOptimizedPaletteForManyBitmaps
//
// -----------------------------------------------------------------------------
  // This separate function is for convenience in processing many bitmaps to
  // obtain an optimized palette.  The CallBack is called until a NIL pointer
  // is returned.
function  CreateOptimizedPaletteforManyBitmaps(const CallBack:  TGetBitmapCallback;
                                                 const ColorBits:  INTEGER): hxPalette;
var
      Bitmap             :  TXBitmap;
      ColorQuantizer     :  TColorQuantizer;
      Counter            :  INTEGER;
      i                  :  INTEGER;
      LogicalPalette     :  TxMaxLogPalette;
      OK                 :  BOOLEAN;
      RGBQuadarray       :  TRGBQuadarray;
      ScreendeviceContext:  hxDC;
begin
    LogicalPalette.palVersion    := PaletteVersion;
    LogicalPalette.palNumEntries := 256;

    ScreendeviceContext := xGetDC(0);
    TRY
      xGetSystemPaletteEntries(ScreendeviceContext,
                                0,256, LogicalPalette.palPalEntry[0]);   // Need first 10 and last 10
    FINALLY
      xReleaseDC(0, ScreendeviceContext)
    end;

    // Normally for 24-bit images, use ColorBits of 5 or 6.  for 8-bit images
    // use ColorBits = 8.
    ColorQuantizer := TColorQuantizer.Create(236, ColorBits);
    TRY

      // Keep calling the callback for more bitmaps until a NIL pointer is
      // returned.  Use Counter and OK parameters is for convenience.
      Counter := 1;
      CallBack(Counter, OK, Bitmap);
      WHILE (Bitmap <> NIL) DO
      begin
        IF   OK
        THEN ColorQuantizer.ProcessImage(Bitmap.Handle);

        INC(Counter);
        CallBack(Counter, OK, Bitmap)
      end;

      ColorQuantizer.GetColorTable(RGBQuadarray);

      // Skip over first 10 and last 10 "fixed" entries
      for i := 0 TO 255-20 DO
      begin
        // Skip over first 10 "fixed" entries in system palette
        LogicalPalette.palPalEntry[10+i].peRed   := RGBQuadarray[i].rgbRed;
        LogicalPalette.palPalEntry[10+i].peGreen := RGBQuadarray[i].rgbGreen;
        LogicalPalette.palPalEntry[10+i].peBlue  := RGBQuadarray[i].rgbBlue;
        LogicalPalette.palPalEntry[10+i].peFlags := RGBQuadarray[i].rgbReserved
      end;
      RESULT := xCreatePalette(pxLogPalette(@LogicalPalette)^)
    FINALLY
      ColorQuantizer.Free
    end
end;

// -----------------------------------------------------------------------------
//
//			doCreateOptimizedPaletteFromSingleBitmap
//
// -----------------------------------------------------------------------------
function doCreateOptimizedPaletteFromSingleBitmap(const DIB: TDIBReader;
  Colors, ColorBits: integer; Windows: boolean): hxPalette;
var
  SystemPalette		: HxPalette;
  ColorQuantizer	: TColorQuantizer;
  i			: integer;
  LogicalPalette	: TxMaxLogPalette;
  RGBQuadArray		: TRGBQuadArray;
  Offset		: integer;
begin
  LogicalPalette.palVersion := $0300;
  LogicalPalette.palNumEntries := Colors;

  if (Windows) then
  begin
    // Get the windows 20 color system palette
    SystemPalette := xGetStockObject(xDEFAULT_PALETTE);
    xGetPaletteEntries(SystemPalette, 0, 10, LogicalPalette.palPalEntry[0]);
    xGetPaletteEntries(SystemPalette, 10, 10, LogicalPalette.palPalEntry[245]);
    Colors := 236;
    Offset := 10;
    LogicalPalette.palNumEntries := 256;
  end else
    Offset := 0;

  // Normally for 24-bit images, use ColorBits of 5 or 6.  For 8-bit images
  // use ColorBits = 8.
  ColorQuantizer := TColorQuantizer.Create(Colors, ColorBits);
  try
    ColorQuantizer.ProcessDIBImage(DIB);
    ColorQuantizer.GetColorTable(RGBQuadArray);
  finally
    ColorQuantizer.Free;
  end;

  for i := 0 to Colors-1 do
    with LogicalPalette.palPalEntry[i+Offset] do
    begin
      peRed   := RGBQuadArray[i].rgbRed;
      peGreen := RGBQuadArray[i].rgbGreen;
      peBlue  := RGBQuadArray[i].rgbBlue;
      peFlags := RGBQuadArray[i].rgbReserved;
    end;
  Result := xCreatePalette(pxLogPalette(@LogicalPalette)^);
end;


function CreateOptimizedPaletteFromSingleBitmap(const Bitmap: TXBitmap;
  Colors, ColorBits: integer; Windows: boolean): hxPalette;
var
  DIB			: TDIBReader;
begin
  DIB := TDIBReader.Create(Bitmap, xpf24bit);
  try
    Result := doCreateOptimizedPaletteFromSingleBitmap(DIB, Colors, ColorBits, Windows);
  finally
    DIB.Free;
  end;
end;





function WebPalette: HxPalette;
type
  TLogWebPalette	= packed record
    palVersion		: word;
    palNumEntries	: word;
    PalEntries		: array[0..5,0..5,0..5] of TxPaletteEntry;
  end;
var
  r, g, b		: byte;
  LogWebPalette		: TLogWebPalette;
  LogPalette		: TxLogpalette absolute LogWebPalette; // Stupid typecast
begin
  with LogWebPalette do
  begin
    palVersion:= $0300;
    palNumEntries:= 216;
    for r:=0 to 5 do
      for g:=0 to 5 do
        for b:=0 to 5 do
        begin
          with PalEntries[r,g,b] do
          begin
            peRed := 51 * r;
            peGreen := 51 * g;
            peBlue := 51 * b;
            peFlags := 0;
          end;
        end;
  end;
  Result := xCreatePalette(Logpalette);
end;



// -----------------------------------------------------------------------------
//
//			ReadPaletteFromFile
//
// -----------------------------------------------------------------------------
function ReadPaletteFromFile(FileName:String):HxPalette;
var F:TextFile;
    s:String;
    i:Integer;
    fLogicalPalette:TxMaxLogPalette;
    fRGBQuadarray:TRGBQuadarray;
begin
  AssignFile(F,FileName);
  Reset(F);
  ReadLn(F,s);
  ReadLn(f,s);
  ReadLn(f,s);
  i:=0;
  while not EOF(f) do
  begin
   ReadLn(f,s);
    fRGBQuadarray[i].rgbRed:=xStrToInt(Copy(s,1,Pos(' ',s)-1));
    Delete(s,1,Pos(' ',s));
    fRGBQuadarray[i].rgbGreen:=xStrToInt(Copy(s,1,Pos(' ',s)-1));
    Delete(s,1,Pos(' ',s));
    fRGBQuadarray[i].rgbBlue:=xStrToInt(s);
    Inc(i,1);
    fRGBQuadarray[i].rgbReserved:=0;
   end;

    fLogicalPalette.palVersion    := PaletteVersion;
    fLogicalPalette.palNumEntries := 256;

      for i := 0 TO 255 DO
      begin
        fLogicalPalette.palPalEntry[i].peRed   := fRGBQuadarray[i].rgbRed;
        fLogicalPalette.palPalEntry[i].peGreen := fRGBQuadarray[i].rgbGreen;
        fLogicalPalette.palPalEntry[i].peBlue  := fRGBQuadarray[i].rgbBlue;
        fLogicalPalette.palPalEntry[i].peFlags := fRGBQuadarray[i].rgbReserved
      end;
      Result := xCreatePalette(pxLogPalette(@fLogicalPalette)^);
end;


// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------
//
//			Apply palettes
//
// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------

// -----------------------------------------------------------------------------
//
//			ApplyPaletteToGraphic
//
// -----------------------------------------------------------------------------
procedure ApplyPaletteToGraphic(Palette:Hxpalette;Graphic:TxGraphic;ChangePixelformat:Boolean;NewPixelformat:TxPixelformat);
begin
  if (Palette<>0) then
  begin
{    if Graphic is TJPEGImage then
    begin
      TJPEGImage(Graphic).Pixelformat:=jf8Bit;
      DeleteObject(TJPEGImage(Graphic).palette);
      TJPEGImage(Graphic).Palette:=Palette;;

    end else if Graphic is TXBitmap then}
    begin
      TXBitmap(Graphic).Pixelformat:=NewPixelformat;
      xDeleteObject(TXBitmap(Graphic).ReleasePalette);
//      TXBitmap(Graphic).Palette:=Palette;
      TXBitmap(Graphic).Palette:=xCopyPalette(Palette);
      TXBitmap(Graphic).Canvas.Refresh;
    end;
  end;
end;

// -----------------------------------------------------------------------------
//
//			ApplyPaletteToDC
//
// -----------------------------------------------------------------------------
procedure ApplyPaletteToDC(DC:HxDC;Palette:HxPalette);
begin
  if Palette<>0 then
  begin
    xSelectPalette(DC, Palette, False);
    xRealizePalette(DC);
  end;
end;


// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------
//
//			Misc. palette procedures
//
// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------

// -----------------------------------------------------------------------------
//
//			LogicalPaletteToRGBQuad
//
// -----------------------------------------------------------------------------
procedure LogicalPaletteToRGBQuad(const Start       :  WORD;
                                    const Count       :  WORD;
                                    var LogicalPalette:  TPaletteEntries;
                                    var RGBQuad       :  TRGBQuadarray);
var
      i:  INTEGER;
begin
    for i := Start TO Start+Count-1 DO
    begin
      RGBQuad[i].rgbRed      := LogicalPalette[i].peRed;
      RGBQuad[i].rgbGreen    := LogicalPalette[i].peGreen;
      RGBQuad[i].rgbBlue     := LogicalPalette[i].peBlue;
      RGBQuad[i].rgbReserved := LogicalPalette[i].peFlags
    end
end;



// -----------------------------------------------------------------------------
//
//			WMQuereyNewPalette
//
// -----------------------------------------------------------------------------
 //  This message informs a window that it is about to receive the
  //  keyboard focus, giving the window the opportunity to realize its
  //  logical palette when it receives the focus.
procedure WMQueryNewPalette(formcanvashandle:TxHandle; PaletteHandle:  hxPalette;
                              var Msg:  TxWMQueryNewPalette);
var
      DeviceContext:  hxDC;
      Palette      :  hxPalette;
      ReturnCode   :  INTEGER;
begin
    DeviceContext := formCanvasHandle;

    // Select the specified palette into a device context.
    // FALSE parameter forces logical palette to be copied into the device palette
    // when the application is in the foreground.
    Palette := xSelectPalette(DeviceContext, PaletteHandle, FALSE);

    // Map palette entries from the current logical palette to the system palette.
    // Returned value is the number of entries in the logical palette mapped to
    // the system palette.
    ReturnCode := xRealizePalette(DeviceContext);

    // Restore the old palette into the device context.
    xSelectPalette(DeviceContext, Palette, FALSE);

    // If new entries were mapped to the system palette, then invalidate the
    // image so it gets repainted.
  //  IF   ReturnCode > 0
  //  THEN form.Invalidate;

    Msg.Result := ReturnCode
end;


// -----------------------------------------------------------------------------
//
//			WMPaletteChanged
//
// -----------------------------------------------------------------------------
  // This message is sent to all top-level and overlapped windows after the
  // window with keyboard focus has realized its logical palette, thereby
  // changing the system palette.  This message enables a window that uses a
  // color palette but does not have the keyboard focus to realize its logical
  // palette and update its client area.
procedure WMPaletteChanged(formcanvashandle:TxHandle;formhandle:TxHandle; PaletteHandle:  hxPalette;
                            var Msg:  TxWMPaletteChanged);
var
      DeviceContext:  hxDC;
      Palette      :  hxPalette;
begin
    IF  Msg.PalChg = formHandle
    THEN begin
      // To avoid creating an infinite loop, a window that receives this message
      // must not realize its palette (unless it determines the window that
      // caused the palette to change is not its own)
      Msg.Result := 0
    end
    else begin
      DeviceContext := formCanvasHandle;

      // Select the specified palette into a device context.
      // A TRUE parameter causes the logical palette to be mapped to the colors
      // already in the physical palette in the best possible way.
      Palette := xSelectPalette(DeviceContext, PaletteHandle, TRUE);

      // Map palette entries from the current logical palette to the system palette.
      // Returned value is the number of entries in the logical palette mapped to
      // the system palette.
      xRealizePalette(DeviceContext);

      // UpdateColors is Microsoft's preferred way to refresh an on-screen Window
      // if the system palette changes.  UpdateColors updates the clinet area
      // of the specified device contenxt by remapping the current colors in the
      // client area to the currently realized logical palette.  An inactive window
      // with a realized logical palette may call UpdateColors as an alternative
      // to redrawing its client area when the system palette changes.
      xUpdateColors(DeviceContext);

      // Restore the old palette into the device context.
      xSelectPalette(DeviceContext, Palette, FALSE);

      Msg.Result := 1;
    end
end;



end.




